/*
   DETH - Doom Editor for Total Headcases, by Simon Oke and Antony Burden.
   HETH - Hexen Editor for Total Headcases, by Antony Burden.
   
   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.  Read the file README.1ST for more information.
   
   This program comes with absolutely no warranty.

   READCFG.C, by Simon Oke.
   */

#include "deu.h"
#include "levels.h"

#include <assert.h>
#include <stdio.h>

#ifdef NULL
#undef NULL
#define NULL ((char *)0x0L)
#endif

struct KVPair
{
	const char *Key;
	Bool IsInt;
	void *Value;
};

BCINT getcol(char *str);
void readcfg(char *file);
void add_thing(char *line);
void add_thing_class(char *line);
void add_spawn(char *line);
void add_spawn_class(char *line);
void add_scheme(char *line);
void add_scheme_class(char *line);
void add_linedef(char *line);
void add_linedef_class(char *line);
void add_sector_type(char *line);
void add_sector_class(char *line);
void add_levels(char *line);
void add_texture_sections(char *line);
void add_ftexture_sections(char *line);
void get_palette_colour(char *line);
void AddKVLine(char *line, struct KVPair *);
void add_key(char *line, key_type **head);
void add_enum_class(char *line);
int read_line(FILE *f, char *buf, int buflen);

/* stuff for manipulating SList s */
SList SList_append(SList l, char *str);
SList SList_find(SList l, char *str);

extern BCINT Palette[];

/* what all the various sections in the cfg file are */
typedef enum
{
	NONE = 0,	/* haven't got a section header yet */
	WAD,
	THINGS,
	SPAWNABLE,
	SPECIALS,
	LEVELS,
	TEXTURES,
	FTEXTURES,
	REGTEST,
	SECTORS,
	COLORS,
	SCHEMES,
	ENUMS,
	FLAGS,
	ICONS,
	DEFAULT_SECTOR,
	DEFAULT_DOOR,
	DEFAULT_STAIR,
	DEFAULT_TELEPORT,
	DEFAULT_TELEPORTDEST,
	DEFAULT_PLAT,
	NUM_SECTIONS
} cfg_section;

char *header[NUM_SECTIONS] =
{
    /* these go hand in hand with the numbers above */
	"",
	"[wad]",
	"[things]",
	"[spawnable]",
	"[specials]",
	"[levels]",
	"[textures]",
	"[ftextures]",
	"[regtest]",
	"[sectors]",
	"[colors]",
	"[schemes]",
	"[enums]",
	"[flags]",
	"[icons]",
	"[default.sector]",
	"[default.door]",
	"[default.stair]",
	"[default.teleport]",
	"[default.teleportdest]",
	"[default.plat]"
};

char *colours[PALETTE_SIZE] =
{
	"BLACK", "BLUE", "GREEN", "CYAN", "RED", "MAGENTA", "BROWN", "LIGHTGRAY",
	"DARKGRAY", "LIGHTBLUE", "LIGHTGREEN", "LIGHTCYAN", "LIGHTRED",
	"LIGHTMAGENTA", "YELLOW", "WHITE", "DARKBLUE", "DARKGREEN", "DARKRED",
	"DARKMAGENTA", "GRAY", "DARKERGRAY", "ORANGE", "SECTORTAGGED", "SECTORSECRET",
	"SECTORPAINFUL", "SECTORLIGHT", "LINEDEFTAGGED", "LINEDEFSECRET",
	"LINEDEFNOSOUND", "LINEDEFNOPASS", "LINEDEFNOMAP", "GFXBLUE", "GFXWADLIST",
	"THINGPATH", "THINGTOPATH", "GRIDCOLOR", "FLOORGRIDCOLOR"
};

ld_class *linedef_class = (ld_class *)NULL;
/* what was the last linedef class we had? */

thing_class *current_thing_class = (thing_class *)NULL;
/* ditto for thing class */

spawn_class *current_spawn_class = (spawn_class *)NULL;
/* ditto for spawn class */

sector_class *current_sector_class = (sector_class *)NULL;
/* ditto for sectors */

scheme_class *current_scheme_class = (scheme_class *)NULL;
/* ditto for texture schemes */

enum_type *current_enum_class = (enum_type *)NULL;

// Keys for [Default.Sector] section
struct KVPair SectorKeys[] =
{
	{ "Above", 0, DefaultUpperTexture },
	{ "Main", 0, DefaultWallTexture },
	{ "Below", 0, DefaultLowerTexture },
	{ "Ceiling", 0, DefaultCeilingTexture },
	{ "Floor", 0, DefaultFloorTexture },
	{ "CeilingHeight", 1, &DefaultCeilingHeight },
	{ "FloorHeight", 1, &DefaultFloorHeight },
	{ "Lighting", 1, &DefaultLighting },
	{ NULL }
};

// Keys for [Default.Door] section
struct KVPair DoorKeys[] =
{
	{ "Base", 0, DefaultDoorBase },
	{ "Door", 0, DefaultDoorTexture },
	{ "Track", 0, DefaultDoorTrack },
	{ NULL }
};

// Keys for [Default.Stair] section
struct KVPair StairKeys[] =
{
	{ "FloorRunner", 0, DefaultLowerStepTexture },
	{ "CeilingRunner", 0, DefaultUpperStepTexture },
	{ NULL }
};

// Keys for [Default.Teleport] section
struct KVPair TeleportKeys[] =
{
	{ "Texture", 0, DefaultTeleport.Texture },
	{ "Decoration", 1, &DefaultTeleport.Decoration },
	{ "Side", 0, DefaultTeleport.Side },
	{ "Top", 1, &DefaultTeleport.Top },
	{ "Bot", 1, &DefaultTeleport.Bottom },
	{ "Light", 1, &DefaultTeleport.Light },
	{ "InsetDepth", 1, &DefaultTeleport.InsetDepth },
	{ "InsetSide", 0, DefaultTeleport.InsetSide },
	{ "InsetRadius", 1, &DefaultTeleport.InsetRadius },
	{ "InsetPoints", 1, &DefaultTeleport.InsetPoints },
	{ NULL }
};

// Keys for [Default.TeleportDest] section
struct KVPair TeleportDestKeys[] =
{
	{ "Texture", 0, DefaultTeleDest.Texture },
	{ "Decoration", 1, &DefaultTeleDest.Decoration },
	{ "Side", 0, DefaultTeleDest.Side },
	{ "Top", 1, &DefaultTeleDest.Top },
	{ "Bot", 1, &DefaultTeleDest.Bottom },
	{ "Light", 1, &DefaultTeleDest.Light },
	{ "InsetDepth", 1, &DefaultTeleDest.InsetDepth },
	{ "InsetSide", 0, DefaultTeleDest.InsetSide },
	{ "InsetRadius", 1, &DefaultTeleDest.InsetRadius },
	{ "InsetPoints", 1, &DefaultTeleDest.InsetPoints },
	{ NULL }
};


// Keys for [Default.Plat] section
struct KVPair PlatKeys[] =
{
	{ "Texture", 0, DefaultPlatTexture },
	{ NULL }
};

BCINT getcol(char *str)
{
	int i = -1;
	
	/* first, do the easy bit -- is the colour specified as a number? */
	if(sscanf(str, "%d", &i))
		return (BCINT)i;
	
	/* check it against the colour names */
	for(i = 0; i < PALETTE_SIZE; i++)
		if(!stricmp(str, colours[i]))
			return (BCINT)i;
	
	return (BCINT)-1;
}




void readcfg(char *file)
{
	static cfg_section what = NONE;
	/* what section are we currently reading? */
	
	FILE *thefile;
	char buf[256], *tok;
	
	int i;
	
	if(!(thefile = fopen(file, "rb")))
		return;
	
	/* OK, we opened the file */
	while(read_line(thefile, buf, 255))
	{
		/* first, see whether it's just a comment */
		if(*buf == '#')
			goto nextline;
		
		if(!strnicmp("include", buf, 7))
		{
			tok = strtok(buf + 7, "\t ");
			readcfg(tok);
			goto nextline;
		}
		
		for(i = 0; i < NUM_SECTIONS; i++)
		{
			if(!stricmp(buf, header[i]))
			{
				what = i;
				goto nextline;
			}
		}
		
		switch(what)
		{
		case NONE:
		case NUM_SECTIONS:
			break;
			
		case WAD:
			MainWad = strdup (buf);
			break;
			
		case THINGS:
			if(buf[0] == '"' && buf[strlen(buf) - 1] ==  '"')
				add_thing_class(buf);
			else
				add_thing(buf);
			break;
			
		case SCHEMES:
			if(buf[0] == '"' && buf[strlen(buf) - 1] ==  '"')
				add_scheme_class(buf);
			else
				add_scheme(buf);
			break;

		case SPAWNABLE:
			if(buf[0] == '"' && buf[strlen(buf) - 1] ==  '"')
				add_spawn_class(buf);
			else
				add_spawn(buf);
			break;

		case SPECIALS:
			if(buf[0] == '"' && buf[strlen(buf) - 1] == '"')
				add_linedef_class(buf);
			else	
				add_linedef(buf);
			break;
			
		case SECTORS:
			if(buf[0] == '"' && buf[strlen(buf) - 1] == '"')
			{
				add_sector_class(buf);
			}
			else if (strlen(buf) && buf[0]!='\n' && buf[0]!='\r')
			{
				add_sector_type(buf);
			}
			break;

		// Flags are loaded just like enums, except they can use
		// their short names for their descriptions.
		case FLAGS:
		case ENUMS:
			if(buf[0] == '"' && buf[strlen(buf) - 1] == '"')
			{
				add_enum_class(buf);
			}
			else if (current_enum_class != (enum_type *)NULL)
			{
				add_key(buf, &current_enum_class->keys);
			}
			break;
			
		case REGTEST:
			RegTest = strdup(buf);
			break;
			
		case LEVELS:
			add_levels(buf);
			break;
			
		case TEXTURES:
			printf("Textures\n");
			add_texture_sections(buf);
			break;
			
		case FTEXTURES:
			add_ftexture_sections(buf);
			break;
			
		case COLORS:
			get_palette_colour(buf);
			break;

		case ICONS:
			OpenIconFile (buf);
			break;

		case DEFAULT_SECTOR:
			AddKVLine (buf, SectorKeys);
			break;

		case DEFAULT_DOOR:
			AddKVLine (buf, DoorKeys);
			break;

		case DEFAULT_STAIR:
			AddKVLine (buf, StairKeys);
			break;

		case DEFAULT_TELEPORT:
			AddKVLine (buf, TeleportKeys);
			break;

		case DEFAULT_TELEPORTDEST:
			AddKVLine (buf, TeleportDestKeys);
			break;

		case DEFAULT_PLAT:
			AddKVLine (buf, PlatKeys);
			break;
		}
		
	nextline:
		;
	}

	fclose(thefile);
}

/* Change the MainWad in a game's .cfg to match the current one */
void fixcfg (char *file)
{
	FILE *out, *in;
	char buff[4096];
	int atWad;
	int someError;
	char *temp;
	
	out = fopen (temp = tmpnam (NULL), "w");
	if (out == 0)
		return;

	in = fopen (file, "r");
	if (in == 0)
	{
		fclose (out);
		return;
	}

	atWad = 0;
	someError = FALSE;

	while (fgets (buff, sizeof(buff), in) != NULL)
	{
		switch (atWad)
		{
		case 0:
			if (stricmp (buff, "[wad]\n") == 0)
			{
				atWad = 1;
			}
			/* intentional fall-through */
		default:
			if (fputs (buff, out) < 0)
			{
				someError = TRUE;
			}
			break;

		case 1:
			if (fprintf (out, "%s\n", MainWad) < 0)
			{
				someError = TRUE;
			}
			atWad = 2;
			break;
		}
	}

	if (!someError)
	{
		remove (file);
		rename (temp, file);
	}
}

void add_key (char *line, key_type **head)
{
	int type, mask = 0;
	char *tok, *endptr, *full, *shortOn = NULL, *shortOff = NULL;
	key_type *new;

	tok = strtok(line, "\t ,");
	if (tok && ((type = strtol(tok, &endptr, 0)) || endptr > tok))
	{
		tok = strtok(NULL, "\t,\"");
		while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
			tok = strtok(NULL, "\t\"");
		full = strdup (tok);

		// get short on name for flags
		tok = strtok(NULL, "\t,\"");
		while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
			tok = strtok(NULL, "\t\"");
		if (tok)
		{
			shortOn = strdup (tok);

			// get short off name for flags
			tok = strtok(NULL, "\t,\"");
			while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
				tok = strtok(NULL, "\t\"");
			if (tok)
			{
				shortOff = strdup (tok);

				// get mask for flags
				tok = strtok(NULL, "\t ,");
				if (tok)
				{
					mask = strtol (tok, &endptr, 0);
				}
			}
		}

		new = (key_type *)myalloc (sizeof(key_type));
		new->type = type;
		new->mask = mask;
		new->name = full;
		new->shortOn = shortOn;
		new->shortOff = shortOff;
		new->next = (key_type *)NULL;

		while (*head)
		{
			head = &(*head)->next;
		}
		*head = new;
	}
}

void add_thing(char *line)
{
	char iconStr[10];
	BCINT col, radius, type, extra = 0, pathtype = 0;
	char *tok, *sprite, *name;
	char args[6] = "     ", *argdesc[5] = { 0 }, *enums[5] = { 0 };
	thing_type *p, *new = (thing_type *)NULL;
	int i;
	
	if(!current_thing_class)
		return;
	
	tok = strtok(line, "\t ,");
	if (tok && stricmp (tok, "delete") == 0)		// Delete an already-defined thing
	{
		thing_class *c;
		thing_type **t;
		tok = strtok (NULL, "\t ,");
		type = strtol (tok, &name, 0);

		for (c = Thing_classes; c; c = c->next)
		{
			t = &c->types;
			while (*t && (*t)->type != type)
			{
				t = &(*t)->next;
			}
			if (*t)
			{
				p = *t;
				*t = p->next;
				free (p);
				return;
			}
		}
	}
	else if(tok && (type = strtol(tok, &name, 0)))	// Add a new thing
	{
		tok = strtok(NULL, "\t ,");
		if(tok && (col = getcol(tok)) != -1)
		{
			tok = strtok(NULL, "\t ,");
			if(tok && (radius = atoi(tok)))
			{

				tok = strtok(NULL, "\t,\"");
				while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
					tok = strtok(NULL, "\t\"");
				name = strdup(tok);
				
				tok = strtok(NULL, "\t,\"");
				while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
					tok = strtok(NULL, "\t\"");
				if (tok && tok[0] == ':')
				{
					sprintf (iconStr, ":%x", TranslateIcon (strtol(&tok[1], &sprite, 0)));
					sprite = strdup (iconStr);
				}
				else
				{
					sprite = strdup(tok);
				}

				// [RH] Parse some possible extra flags for things
				tok = strtok(NULL, "\t ,");
				while (tok && *tok != '\"')
				{
					if (stricmp (tok, "Poly") == 0)
					{
						extra |= TX_POLYSPOT;
					}
					else if (stricmp (tok, "DeathScript") == 0)
					{
						extra |= TX_DEATHSCRIPT;
					}
					else if (stricmp (tok, "NoAngle") == 0)
					{
						extra |= TX_NOANGLE;
					}
					else if (stricmp (tok, "Start") == 0)
					{
						NumSingleStarts++;
						extra |= TX_START;
					}
					else if (stricmp (tok, "DMStart") == 0)
					{
						extra |= TX_DMSTART;
					}
					else if (stricmp (tok, "SectorAction") == 0)
					{
						extra |= TX_SECTORACTION;
					}
					else if (stricmp (tok, "Resizer") == 0)
					{
						extra |= TX_RESIZER;
					}
					else if (stricmp (tok, "Path") == 0)
					{
						extra |= TX_PATH;
					}
					else if (strnicmp (tok, "ToPath=",7) == 0)
					{
						extra |= TX_TOPATH;
						pathtype = strtol (tok+7, &tok, 0);
					}
					else if (stricmp (tok, "NoPic") == 0)
					{
						extra |= TX_NOPIC;
					}
					tok = strtok(NULL, "\t ,");
				}

				// [RH] Sprites can use their arguments for purposes other than specials
				if (tok && *tok == '\"')
				{
					strncpy (args, tok+1, 5);
					tok = strtok(NULL, "\t,\"");
					if (tok)
					{
						for (i = 0; i < 5; ++i)
						{
							while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
								tok = strtok(NULL, "\t\"");
							argdesc[i] = strdup (tok);
							tok = strtok(NULL, "\t\"");
						}

						// [RH] Get names of parameter lists
						for (i = 0; i < 5; ++i)
						{
							if (args[i] != 'E' && args[i] != 'F')
								continue;

							while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
								tok = strtok(NULL, "\t\"");
							enums[i] = strdup (tok);
							tok = strtok(NULL, "\t\"");
						}
					}
				}


				/* do the add here */
				for (p = current_thing_class->types; p && p->next; p = p->next)
				{
				}
				
				new = (thing_type *)myalloc(sizeof(thing_type));
				assert(new);
				
				new->type = type;
				new->colour = col;
				new->radius = radius;
				new->extra = extra;
				new->name = name;
				new->sprite = sprite;
				memcpy (new->args, args, sizeof(args));
				new->pathtype = pathtype;
				memcpy (new->argtext, argdesc, sizeof(argdesc));
				memcpy (new->enums, enums, sizeof(enums));
				new->next = (thing_type *)NULL;

				if(p)
					p->next = new;
				else
					current_thing_class->types = new;
			}
		}
	}
}

void add_thing_class(char *line)
{
	thing_class *p, *new = (thing_class *)NULL;
	
	line[strlen(line) - 1] = '\0';
	line++;
	/* add it here */
	
	for(p = Thing_classes; (p && p->next); p = p->next)
		if(!strcmp(line, p->name)) {
			current_thing_class = p;
			return;
		}
	
	new = (thing_class *)myalloc(sizeof(thing_class));
	assert(new);
	
	new->name = strdup(line);
	new->next = (thing_class *)NULL;
	new->types = (thing_type *)NULL;		/*JFF initialize this pointer*/
	
	if(p)
		p->next = new;
	else
		Thing_classes = new;
	
	current_thing_class = new; 
}

void add_enum_class(char *line)
{
	enum_type *p, *newp = (enum_type *)NULL;

	line[strlen(line) - 1] = '\0';
	line++;

	for (p = Enum_classes; p && p->next; p = p->next)
	{
		if (strcmp (line, p->name) == 0)
		{
			current_enum_class = p;
			return;
		}
	}

	newp = (enum_type *)myalloc(sizeof(enum_type));
	newp->name = strdup (line);
	newp->next = (enum_type *)NULL;
	newp->count = 0;
	newp->array = (key_ptr *)NULL;
	newp->keys = (key_type *)NULL;

	if (p)
		p->next = newp;
	else
		Enum_classes = newp;

	current_enum_class = newp;
}

void add_spawn(char *line)
{
	BCINT type;
	char *tok, *sprite, *name;
	spawn_type *p, *new = (spawn_type *)NULL;
	
	if(!current_spawn_class)
		return;

	tok = strtok(line, "\t ,");
	if(tok && (type = atoi(tok)) != -1)
	{
/*	if(tok && (type = atoi(tok))) { */
		tok = strtok(NULL, "\t,\"");
		while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
			tok = strtok(NULL, "\t\"");
		name = strdup(tok);
		
		tok = strtok(NULL, "\t,\"");
		while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
			tok = strtok(NULL, "\t\"");
		sprite = strdup(tok);
		

		/* do the add here */
		for(p = current_spawn_class->types;
			(p && p->next); p = p->next);

		new = (spawn_type *)myalloc(sizeof(spawn_type));
		assert(new);
		
		new->type = type;
		new->name = name;
		new->sprite = sprite;
		new->next = (spawn_type *)NULL;

		if(p)
			p->next = new;
		else
			current_spawn_class->types = new;
	}
}

void add_spawn_class(char *line)
{
	spawn_class *p, *new = (spawn_class *)NULL;
	
	line[strlen(line) - 1] = '\0';
	line++;
	/* add it here */
	
	for(p = Spawn_classes; (p && p->next); p = p->next)
		if(!strcmp(line, p->name)) {
			current_spawn_class = p;
			return;
		}
	
	new = (spawn_class *)myalloc(sizeof(spawn_class));
	assert(new);
	
	new->name = strdup(line);
	new->next = (spawn_class *)NULL;
	new->types = (spawn_type *)NULL;		/*JFF initialize this pointer*/
	
	if(p)
		p->next = new;
	else
		Spawn_classes = new;
	
	current_spawn_class = new; 
}



void add_linedef(char *line)
{
	char *short_desc, *long_desc, *tok, *argdesc[5], *enums[5] = { 0 };
	char args[6] = "-----";
	BCINT type = -1;
	ld_type *p, *new = (ld_type *)NULL;
	int i;

	if(!linedef_class)
		return;
	/* can't add a linedef if we don't know what class
	   to put it in */

	tok = strtok(line, "\t ,");
	if(tok && (type = atoi(tok)) != -1)
	{
		tok = strtok(NULL, "\t,\"");
		while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
			tok = strtok(NULL, "\t\"");
		/* should by now have the short description */
		short_desc = strdup(tok);
		
		tok = strtok(NULL, "\t,\"");
		while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
			tok = strtok(NULL, "\t\"");
		/* should by now have the long description */
		long_desc = strdup(tok);
		
		tok = strtok(NULL, "\t,\"");
		while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
			tok = strtok(NULL, "\t\"");
		/* should by now have the argument types */
		strncpy (args, tok, 5);

		for (i = 0; i < 5; ++i)
		{
			tok = strtok(NULL, "\t,\"");
			while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
				tok = strtok(NULL, "\t\"");
			/* should by now have argument i */
			argdesc[i] = strdup(tok);
		}

		// [RH] Get names of parameter lists
		for (i = 0; i < 5; ++i)
		{
			if (args[i] != 'E' && args[i] != 'F')
				continue;

			tok = strtok(NULL, "\t,\"");
			while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
				tok = strtok(NULL, "\t\"");
			enums[i] = strdup(tok);
		}

		/* now add it */
		for(p = linedef_class->types; (p && p->next); p = p->next);
		/* make p point to the last item in the list, or be NULL */

		new = (ld_type *)myalloc(sizeof(ld_type));
		assert(new);

		new->longname = long_desc;
		new->shortname = short_desc;
		memcpy (new->args, args, sizeof(args));
		memcpy (new->argtext, argdesc, sizeof(argdesc));
		memcpy (new->enums, enums, sizeof(enums));
		new->type = type;
		new->next = (ld_type *)NULL;

		if(p)
			p->next = new;
		else
			linedef_class->types = new;
	}
}

void add_linedef_class(char *line)
{
	ld_class *p, *new = (ld_class *)NULL;
	
	line++;
	line[strlen(line) - 1] = '\0';
	/* lose the quotes */
	
	for(p = Linedef_classes; (p && p->next); p = p->next)
		if(!strcmp(p->name, line)) {
			linedef_class = p;
			return;
			/* we already had it */
		}
	
	/* p now points to the last item in the list, or is NULL
	   if there is no list. */
	
	new = (ld_class *)myalloc(sizeof(ld_class));
	assert(new);
	/* get memory */
	
	new->name = strdup(line);
	new->types = (ld_type *)NULL;
	new->next = (ld_class *)NULL;
	/* fill in the fields */
	
	if(p)
		p->next = new;
	else
		Linedef_classes = new;
	
	linedef_class = new;
}

void add_scheme(char *line)
{
	BCINT type;
	char *tok, *name, *normal, *upper, *lower, *floor, *ceiling;
	scheme_type *p, *new = (scheme_type *)NULL;
	
	if(!current_scheme_class)
		return;

	tok = strtok(line, "\t ,");
	if(tok && (type = atoi(tok)) != -1) {
	
/*	if(tok && (type = atoi(tok))) { */

		tok = strtok(NULL, "\t,\"");
		while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
			tok = strtok(NULL, "\t\"");
		name = strdup(tok);
		
		tok = strtok(NULL, "\t,\"");
		while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
			tok = strtok(NULL, "\t\"");
		normal = strdup(tok);
		
		tok = strtok(NULL, "\t,\"");
		while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
			tok = strtok(NULL, "\t\"");
		upper = strdup(tok);

		tok = strtok(NULL, "\t,\"");
		while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
			tok = strtok(NULL, "\t\"");
		lower = strdup(tok);

		tok = strtok(NULL, "\t,\"");
		while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
			tok = strtok(NULL, "\t\"");
		floor = strdup(tok);

		tok = strtok(NULL, "\t,\"");
		while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
			tok = strtok(NULL, "\t\"");
		ceiling = strdup(tok);


		/* do the add here */
		for(p = current_scheme_class->types;
			(p && p->next); p = p->next);
		
		new = (scheme_type *)myalloc(sizeof(scheme_type));
		assert(new);
		
		new->type = type;
		new->name = name;
		new->normal = normal;
		new->upper = upper;
		new->lower = lower;
		new->floor = floor;
		new->ceiling = ceiling;

		new->next = (scheme_type *)NULL;
		
		if(p)
			p->next = new;
		else
			current_scheme_class->types = new;
	}
}

void add_scheme_class(char *line)
{
	scheme_class *p, *new = (scheme_class *)NULL;
	
	line[strlen(line) - 1] = '\0';
	line++;
	/* add it here */
	
	for(p = Scheme_classes; (p && p->next); p = p->next)
		if(!strcmp(line, p->name)) {
			current_scheme_class = p;
			return;
		}
	
	new = (scheme_class *)myalloc(sizeof(scheme_class));
	assert(new);
	
	new->name = strdup(line);
	new->next = (scheme_class *)NULL;
	new->types = (scheme_type *)NULL;		/*JFF initialize this pointer*/
	
	if(p)
		p->next = new;
	else
		Scheme_classes = new;
	
	current_scheme_class = new; 
}

void AddKVLine(char *line, struct KVPair *keys)
{
	char *tok = strtok (line, "\t ,=");
	if (tok)
	{
		while (keys->Key != NULL)
		{
			if (stricmp (keys->Key, tok) == 0)
			{
				tok = strtok (NULL, "\t ,=");
				if (keys->IsInt)
				{
					*((BCINT *)keys->Value) = strtol (tok, &line, 0);
				}
				else
				{
					strncpy (keys->Value, tok, 8);
				}
				return;
			}
			++keys;
		}
	}
}


void add_sector_class(char *line)
{
	sector_class *p, *new = (sector_class *)NULL;
	
	line[strlen(line) - 1] = '\0';
	line++;
	
	for(p = Sector_classes; (p && p->next); p = p->next)
		if(!strcmp(line, p->name)) {
			current_sector_class = p;
			return;
		}
	
	new = (sector_class *)myalloc(sizeof(sector_class));
	assert(new);
	
	new->name = strdup(line);
	new->next = (sector_class *)NULL;
	new->types = (sector_type *)NULL;		/*JFF initialize this pointer*/
	
	if(p)
		p->next = new;
	else
		Sector_classes = new;
	
	current_sector_class = new; 
}

void add_sector_type(char *line)
{
	char *short_desc;
	char *long_desc;
	BCINT type = -1;
	char *tok;
	sector_type *p, *new = (sector_type *)NULL;
	
	tok = strtok(line, "\t ,");
	if(tok && (type = atoi(tok)) != -1) {
		tok = strtok(NULL, "\t\"");
		while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
			tok = strtok(NULL, "\t\"");
		/* should by now have the short description */
		short_desc = strdup(tok);
		
		tok = strtok(NULL, "\t\"");
		while(tok && ((tok[0] == ' ') || (tok[0] == ',')))
			tok = strtok(NULL, "\t\"");
		/* should by now have the long description */
		long_desc = strdup(tok);
		
		/* add it now */
		new = (sector_type *)myalloc(sizeof(sector_type));
		assert(new);
		new->shortname = short_desc;
		new->longname = long_desc;
		new->type = type;
		new->next = (sector_type *)NULL;
		
		for(p = current_sector_class->types; (p && p->next); p = p->next);
		
		if(p)
			p->next = new;
		else
			current_sector_class->types = new;
	}
}

void add_levels(char *line)
{
	char *tok;
	
	tok = strtok(line, "\t ,");

	if(tok && !stricmp(tok, "format")) {
		tok = strtok(NULL, "\t ,");
		while(tok) {
			LevelNameFormat = SList_append(LevelNameFormat, tok);
			tok = strtok(NULL, "\t ,");
		}
	}
	else {
		while(tok) {
			/* add a level */
			LevelNames = SList_append(LevelNames, tok);
			
			tok = strtok(NULL, "\t ,");
		}
	}
}

void add_texture_sections(char *line)
{
	char *tok;
	
	tok = strtok(line, "\t ,");
	while(tok) {
		/* add it to list */
		Texture_sections = SList_append(Texture_sections, tok);
		
		tok = strtok(NULL, "\t ,");
	}
}

void add_ftexture_sections(char *line)
{
	char *tok;
	
	tok = strtok(line, "\t ,");
	while(tok) {
		/* add it to list */
		Ftexture_sections = SList_append(Ftexture_sections, tok);
		
		tok = strtok(NULL, "\t ,");
	}
}

void get_palette_colour(char *line)
{
	char *tok;
	BCINT color = -1;
	BCINT palette_index;
	
	tok = strtok(line, "\t ,=");
	if(tok) {
		color = getcol(tok);
		if(color == -1)
			return;
		
		tok = strtok(NULL, "\t ,=");
		palette_index = atoi(tok);
		
		Palette[color] = palette_index;
	}
}



/* read a line in from the current file
   return zero if we hit the end of the file */
int read_line(FILE *f, char *buf, int buflen)
{
	int eof = 0, eol = 0;
	int c;
	
	while(((c = fgetc(f)) == '\n') ||
		  (c == '\r') ||
		  (c == ' ') ||
		  (c == '\t'));
	/* ignore leading whitespace */
	
	while(!(eof || eol) && (buflen > 0)) {
		if(c == EOF)
			eof = 1;
		else if((c == '\n') || (c == '\r'))
			eol = 1;
		else {
			*buf++ = (char)c;
			buflen--;
            c = fgetc(f);
		}
	}

	while ((c = *(buf - 1), c == ' ') || (c == '\t'))
	{
		buf--;
	/* remove trailing whitespace */
	}
	
	*buf = 0; /* bung a terminator on it */

	return(1 - eof);
}

SList SList_append(SList l, char *str)
{
	SList nw, head = l;
	
	while (l && l->next)	/*JFF prevent dereference of NULL pointer*/
		l = l->next;
	
	nw = (SList)myalloc(sizeof(struct _SList));
	if(head)
		l->next = nw;
	else
		head = nw;
	
	nw->string = strdup(str);
	nw->next = (SList)NULL;
	
	return head;
}

SList SList_find(SList l, char *str)
{
	while(l)
		if(!strcmp(l->string, str))
			return l;
		else
			l = l->next;
	
	return (SList)NULL;
}

